using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Policy;
using Leap;
using MacGruber;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class FlexRegion : MonoBehaviour
    {
        private RBDynamics[] rbDynamics;
        private RBDynamics rbDynamicAlt;
        public List<FlexRegionSupportTracker> supportTrackers = new List<FlexRegionSupportTracker>();
        public int id;
        public string name;
        public List<List<DAZMorph>> morphs = new List<List<DAZMorph>>();
        private List<List<MorphBuffer>> morphBuffers = new List<List<MorphBuffer>>();
        private bool[] doDyn = {true,true,true,true};
        private bool[] doStatic = {true,true,true,true};
        public bool supported;
        public FlexRegion[] children;
        
        public HeelTrigger lHeelTrigger;
        public HeelTrigger rHeelTrigger;
        public HeelTrigger lToeTrigger;
        public HeelTrigger rToeTrigger;
        
        public int morphsCount => morphs.Count;

        public JSONStorableBool enabledJ = new JSONStorableBool("Enabled", true);
        public JSONStorableVector4 dynamicStrengthJ = new JSONStorableVector4(
            "Dynamic",
            new Vector4(1f, 1f, 1f, 1f),
            new Vector4(),
            new Vector4(10f,10f,10f, 10f));
        public JSONStorableVector4 amountJ = new JSONStorableVector4(
            "Amount",
            new Vector4(2f, 2f, 2f, 2f),
            new Vector4(),
            new Vector4(5f,5f,5f, 5f));
        public JSONStorableVector4 minAccelJ = new JSONStorableVector4(
            "minAcc",
            new Vector4(0f, 0f, 0f, 0f),
            new Vector4(),
            new Vector4(20f,20f,20f, 20f));
        
        public JSONStorableVector4 sRatioJ = new JSONStorableVector4(
            "S-Ratio",
            new Vector4(.75f,.75f,.75f, .75f),
            new Vector4(0f, 0f, 0f, 0f),
            new Vector4(1f,1f,1f, 1f));
        
        private Dictionary<int, string> sides = new Dictionary<int, string>
        {
            { 0, "L" }, { 1, "R" }
        };

        private Vector4 sRatio { get; set; }
        private Vector4 dynamicStrength { get; set; }
        private Vector4 staticStrength { get; set; }
        private Vector4 minAccel { get; set; }
        // private Vector4 maxAccel { get; set; }
        private Vector4 amount { get; set; }

        private Func<int, int, float> GetAcceleration;

        private void OnEnable()
        {
            supportTrackers.ForEach(x => x.SetActive(true));
        }

        public void Init(int id, string region, string[] rbNames, bool sym = true)
        {
            this.id = id;
            name = region;
            var atom = gameObject.GetAtom();
            Rigidbody rb = null;
            if (sym)
            {
                rbDynamics = new RBDynamics[2];
                for (int i = 0; i < 2; i++)
                {
                    string side = sides[i];
                    rb = atom.forceReceivers.First(x => x.name == side.ToLower() + rbNames[0])
                        .GetComponent<Rigidbody>();
                    rbDynamics[i] = rb.AddRbDynamic(i);
                    SetupSupportTrackers(rb);
                }
            }
            else
            {
                rbDynamics = new RBDynamics[1];
                rb = atom.forceReceivers.First(x => x.name == rbNames[0])
                    .GetComponent<Rigidbody>();
                rbDynamics[0] = rb.AddRbDynamic();
            }

            if (rbNames.Length == 2)
            {
                rb = atom.forceReceivers.First(x => x.name == rbNames[1])
                    .GetComponent<Rigidbody>();
                rbDynamicAlt = rb.AddRbDynamic();
                GetAcceleration = (directionId, sideId) =>
                {
                    float val;
                    if (directionId < 2) val = rbDynamics[sideId].angularAcceleration.x - .75f*rbDynamicAlt.angularAcceleration.x;
                    else val = rbDynamics[sideId].angularAcceleration.z - .75f*rbDynamicAlt.angularAcceleration.z;
                    val *= amount[directionId];
                    return val;
                };
            }
            else
            {
                if(id == 4)
                    GetAcceleration = (directionId, sideId) =>
                    {
                        float val;
                        if (directionId < 2) val = -rbDynamics[sideId].angularAcceleration.y;
                        else val = -rbDynamics[sideId].angularAcceleration.z;
                        val *= amount[directionId];
                        return val;
                    };
                else
                    GetAcceleration = (directionId, sideId) =>
                    {
                        float val;
                        if (directionId < 2) val = rbDynamics[sideId].angularAcceleration.x;
                        else val = rbDynamics[sideId].angularAcceleration.z;
                        val *= amount[directionId];
                        return val;
                    };
            }

            for (int i = 0; i < 4; i++)
            {
                bool morphsFound = false;
                List<DAZMorph> sideMorphs = new List<DAZMorph>();
                if(sym)
                {
                    for (int j = 0; j < 2; j++)
                    {
                        string side = sides[j];
                        var morph = FlexIt.GetMorphByUid($"FI_{region}{i}{side}");
                        if (morph == null) continue;
                        morph.jsonFloat.constrained = false;
                        sideMorphs.Add(morph);
                        morphsFound = true;
                    }
                }
                else
                {
                    var morph = FlexIt.GetMorphByUid($"FI_{region}{i}");
                    if (morph == null) continue;
                    morph.jsonFloat.constrained = false;
                    sideMorphs.Add(morph);
                    morphsFound = true;
                }
                if (!morphsFound) continue;
                morphs.Add(sideMorphs);
                morphBuffers.Add(sideMorphs.Select(x => new MorphBuffer(x)).ToList());
            }

            enabledJ.setCallbackFunction += val => enabled = val;
            dynamicStrengthJ.setCallbackFunction += val => OnDynamicStrengthChanged();
            minAccelJ.setCallbackFunction += val => minAccel = val;
            sRatioJ.setCallbackFunction += OnSRatioChanged;
            amountJ.setCallbackFunction += OnAmountChanged;
            SetupDefaults(id);
            
        }

        private void SetupSupportTrackers(Rigidbody rb)
        {
            List<FlexCollisionTrigger> triggers = new List<FlexCollisionTrigger>();
            if (id == 0)
            {
                string side = rb.name.Substring(0, 1);
                if (side == "l")
                {
                    lHeelTrigger = rb.transform.Find("lHeelColliderBase/lHeelColTOffset/lHeelColROffset/lHeelCollider/lHeelColliderTrigger")?.gameObject.AddComponent<HeelTrigger>();
                    lToeTrigger = rb.transform.Find("lToe/lToeColliderBase/lToeColTOffset/lToeColROffset/lToeCollider/lToeColliderTrigger")?.gameObject.AddComponent<HeelTrigger>();
                }
                else
                {
                    rHeelTrigger = rb.transform.Find("rHeelColliderBase/rHeelColTOffset/rHeelColROffset/rHeelCollider/rHeelColliderTrigger")?.gameObject.AddComponent<HeelTrigger>();
                    rToeTrigger = rb.transform.Find("rToe/rToeColliderBase/rToeColTOffset/rToeColROffset/rToeCollider/rToeColliderTrigger")?.gameObject.AddComponent<HeelTrigger>();
                }
                var person = gameObject.GetAtom();
                triggers.Add(rb.gameObject.AddComponent<FlexCollisionTrigger>());
                var toe = person.rigidbodies.First(x => x.name == side + "Toe");
                var bigToe = person.rigidbodies.First(x => x.name == side + "BigToe");
                var smallToe1 = person.rigidbodies.First(x => x.name == side + "SmallToe1");
                triggers.Add(bigToe.gameObject.AddComponent<FlexCollisionTrigger>());
                triggers.Add(toe.gameObject.AddComponent<FlexCollisionTrigger>());
                var tracker = new FlexRegionSupportTracker(this, rb, triggers, smallToe1);
                supportTrackers.Add(tracker);
            }
            else if (id == 1 || id == 2)
            {
                var colliders = rb.transform.GetChild(3).transform.GetChildren().Select(x => 
                    x.GetChild(0).GetComponent<Collider>()).ToList();
                colliders.Add(rb.GetComponentInChildren<Collider>());
                for (int i = 0; i < colliders.Count; i++)
                {
                    var trigger = colliders[i].gameObject.AddComponent<FlexCollisionTrigger>();
                    triggers.Add(trigger);
                }
                if (triggers.Count > 0)
                {
                    supportTrackers.Add(new FlexRegionSupportTracker(this, rb, triggers));
                }
            }
        }

        public void UpdateSupportTrackers()
        {
            supportTrackers.ForEach(x => x.Destroy());
            supportTrackers.Clear();
            foreach (var rbDynamic in rbDynamics)
            {
                SetupSupportTrackers(rbDynamic.rigidbody);
            }
        }
        
        private void OnAmountChanged(Vector4 val)
        {
            amount = val;
            OnDynamicStrengthChanged();
            staticStrength = MultiplyV4(sRatio, amount);
            for (int i = 0; i < morphBuffers.Count; i++)
            {
                for(int j=0; j<morphBuffers[i].Count; j++) morphBuffers[i][j].max = amount[i];
            }
        }

        private void OnSRatioChanged(Vector4 val)
        {
            sRatio = val;
            doDyn = new[] {val[0] < 1f, val[1] < 1f, val[2] < 1f, val[3] < 1f};
            doStatic = new[] {val[0] > 0f, val[1] > 0f, val[2] > 0f, val[3] > 0f};
            staticStrength = MultiplyV4(sRatio, amount);
        }

        private void OnDynamicStrengthChanged()
        {
            dynamicStrength = MultiplyV4(dynamicStrengthJ.val, amountJ.val);
        }

        private void SetupDefaults(int id)
        {
            Vector4 dynamicD = dynamicStrengthJ.defaultVal;
            Vector4 minAccelD = minAccelJ.defaultVal;
            Vector4 sRatioD = sRatioJ.defaultVal;
            Vector4 amountD = amountJ.defaultVal;
            switch (id)
            {
                case 0:
                    amountD.x = 1.5f;
                    amountD.y = 2f;
                    dynamicD.x = .25f;
                    dynamicD.y = .2f;
                    break;
                case 1:
                    sRatioD.x = sRatioD.y =.75f;
                    dynamicD.y = .75f;
                    // dynamicD.x = 0f;
                    // amountD.x = 5f;
                    break;
                case 2:
                    amountD.x = amountD.x = 1f;
                    dynamicD.x = 2f;
                    dynamicD.y = .25f;
                    break;
                case 3:
                    dynamicD = new Vector4(.35f, .125f, .5f, .5f);
                    sRatioD.x = .5f;
                    break;
                case 4:
                    amountD.w = 1f;
                    dynamicD = new Vector4(1.2f, 1.2f, 1.2f, 1.2f);
                    break;
                case 5:
                    amountD.x = 1f;
                    amountD.y = .65f;
                    break;
            }
            amountJ.val = amountJ.defaultVal = amountD;
            sRatioJ.val = sRatioJ.defaultVal = sRatioD;
            dynamicStrengthJ.val = dynamicStrengthJ.defaultVal = dynamicD;
            minAccel = minAccelJ.val = minAccelJ.defaultVal = minAccelD;
            
            OnAmountChanged(amountJ.val);
            OnSRatioChanged(sRatioJ.val);
        }

        public void FixedUpdate()
        {
            float speed = FlexIt.flexRegressionSpeed;
            for (int i = 0; i < morphs.Count; i++)
            {
                if(amount[i] == 0f) continue;
                bool dyn = doDyn[i];
                bool stat = doStatic[i];
                for (int j = 0; j < morphs[i].Count; j++)
                {
                    morphs[i][j].max = amount[i];
                    float mVal = morphs[i][j].morphValue;
                    float mTarget = 0f;
                    bool flexed = false;
                    float flex = 0f;
                    if(stat)
                    {
                        if((supportTrackers.Count == 0 || !supportTrackers[j].supported) && 
                           (children == null || children.All(x => !x.supportTrackers[j].supported)))
                        {
                            if (i < 2)
                            {
                                if (id == 0)
                                {
                                    flex = FlexIt.MapRange(rbDynamics[j].rigidbody.transform.localRotation.x,
                                        .25f, -.15f, .7f, 1f, -1f);
                                }
                                else
                                {
                                    flex = rbDynamics[j].rigidbody.transform.forward.y;
                                }

                                flexed = i == 0 ? flex > 0f : flex < 0f;
                            }
                            else
                            {
                                flex = rbDynamics[j].rigidbody.transform.right.y;
                                flexed = i == 2 ? flex > 0f : flex < 0f;
                                if (j == 1) flexed = !flexed;
                            }
                        }
                    }
                    if(flexed)
                    {
                        flex = Mathf.Abs(flex);
                        flex = ShakeIt.sin[(int)(flex * 90)];
                        mTarget = flex * staticStrength[i];
                    }
                    if(Mathf.Abs(mVal - mTarget) > .02f)
                    {
                        morphs[i][j].morphValue = Mathf.Lerp(mVal, mTarget, speed);
                    }
                    if(dyn)
                    {
                        
                        float f = GetAcceleration(i, j);
                        bool inRange = (i == 1 || i == 3) ? f > minAccel[i] : f < -minAccel[i];
                        if (inRange)
                        {
                            morphBuffers[i][j].Fill(Mathf.Abs(f) * dynamicStrength[i] * FlexIt.realFrametime);
                        }
                        morphBuffers[i][j].Pop();
                    }
                }
            }
        }

        public void Reset()
        {
            
            for (int i = 0; i < morphs.Count; i++)
            {
                for(int j=0; j<morphs[i].Count; j++) morphs[i][j].morphValue = 0f;
            }
        }

        public void OnDisable()
        {
            // for (int i = 0; i < rbDynamics.Length; i++) rbDynamics[i].uses--;
            supportTrackers.ForEach(x => x.SetActive(false));
            Reset();
        }

        public void OnDestroy()
        {
            supportTrackers.ForEach(x => x.Destroy());
        }

        private Vector4 MultiplyV4(Vector4 a, Vector4 b)
        {
            for (int i = 0; i < 4; i++) a[i] *= b[i];
            return a;
        }
        
        public void Store(JSONClass jc, bool forceStore = false)
        {
            JSONClass tempJC = new JSONClass();
            enabledJ.Store(tempJC, true);
            amountJ.Store(tempJC, true);
            sRatioJ.Store(tempJC, true);
            dynamicStrengthJ.Store(tempJC, true);
            jc[name] = tempJC;
        }

        public void Load(JSONClass jc)
        {
            if (jc.HasKey(name))
            {
                jc = jc[name].AsObject;
                enabledJ.Load(jc);
                amountJ.Load(jc);
                sRatioJ.Load(jc);
                dynamicStrengthJ.Load(jc);
            }
        }
    }

    public class MorphBuffer
    {
        private DAZMorph morph;
        private float buffer;
        public float max;
        public static float increment = .25f;

        public MorphBuffer(DAZMorph morph)
        {
            this.morph = morph;
        }

        public void Fill(float val)
        {
            buffer += val;
        }

        public void Pop()
        {
            if (buffer < .001f) return;
            float val = buffer * increment;
            buffer -= val;
            float newVal = morph.morphValue + val;
            morph.morphValue = newVal > max ? max : newVal;
        }
    }
}